""" python api functions """
import os
import glob
import time
import requests
from labii_sdk.api_client import APIObject

class LabiiObject:
    """ object for labii sdk """
    api = None

    def __init__(self,
        organization__sid,
        base_url="https://www.labii.dev",
        email=None,
        password=None
    ):
        self.api = APIObject(base_url=base_url, email=email, password=password, organization__sid=organization__sid)
        # organizations
        self.Organization = self.APIResource(self, "organizations", "organization")
        self.People = self.APIResource(self, "organizations", "personnel")
        self.Team = self.APIResource(self, "organizations", "team")
        self.OrganizationWidget = self.APIResource(self, "organizations", "organizationwidget")
        self.SAML = self.APIResource(self, "organizations", "saml")
        self.Backup = self.APIResource(self, "organizations", "backup")
        self.Invoice = self.APIResource(self, "organizations", "invoice")
        self.Subscription = self.APIResource(self, "organizations", "subscription")
        self.Credit = self.APIResource(self, "organizations", "credit")
        self.Seat = self.APIResource(self, "organizations", "seat")
        # projects
        self.Project = self.APIResource(self, "projects", "project")
        self.Member = self.APIResource(self, "projects", "member")
        # tables
        self.Application = self.APIResource(self, "organizations", "application")
        self.Table = self.APIResource(self, "tables", "table")
        self.Column = self.APIResource(self, "tables", "column")
        self.Section = self.APIResource(self, "tables", "section")
        self.Filter = self.APIResource(self, "tables", "filter")
        self.Record = self.APIResource(self, "tables", "row")
        self.Cell = self.APIResource(self, "tables", "cell")
        self.Signer = self.APIResource(self, "tables", "signer")
        self.Version = self.APIResource(self, "tables", "version")
        self.Visitor = self.APIResource(self, "tables", "visitor")
        self.Activity = self.APIResource(self, "tables", "activity")
        self.Permission = self.APIResource(self, "tables", "permission")
        # widget
        self.Widget = self.APIResource(self, "widgets", "widget")
        # workflow
        self.Workflow = self.APIResource(self, "workflows", "workflow")

    class APIResource:
        """ abstract class """
        app = None
        model = None

        class Meta:
            """ meta """
            abstract = True

        def __init__(self, instance, app, model):
            """
                - instance, the outer instance
            """
            self.instance = instance
            self.app = app
            self.model = model

        def create(self, data, query=""):
            """
                Create a object
                Args:
                - data (dict), the object data
            """
            return self.instance.api.post(
                self.instance.api.get_list_url(self.app, self.model, serializer="detail", query=query),
                data
            )

        def retrieve(self, sid, query=""):
            """ Return an object """
            return self.instance.api.get(self.instance.api.get_detail_url(self.app, self.model, sid=sid, query=query))

        def list(self, page=1, page_size=10, all_pages=False, level="organization", serializer="list", query=""):
            """ Return list of objects """
            if all_pages is True:
                url = self.instance.api.get_list_url(
                    self.app,
                    self.model,
                    sid=self.instance.api.organization__sid,
                    level=level,
                    serializer=serializer,
                    query=query
                )
                return self.instance.api.get(url, True)
            # not all pages
            url = self.instance.api.get_list_url(
                self.app,
                self.model,
                sid=self.instance.api.organization__sid,
                level=level,
                serializer=serializer,
                query=f"page={page}&page_size={page_size}{'' if query == '' else '&'}{query}"
            )
            return self.instance.api.get(url)

        def modify(self, sid, data, query=""):
            """
                Change one object
                Args:
                - data (dict), the object data
            """
            return self.instance.api.patch(
                self.instance.api.get_detail_url(self.app, self.model, sid=sid, query=query),
                data
            )

        def delete(self, sid, query=""):
            """ Delete a object """
            self.instance.api.delete(self.instance.api.get_detail_url(self.app, self.model, sid=sid, query=query))

        def upload(self, sid, filename):
            """
                Upload a file for files table
                Based on files/FileUpload.js
                Need a file row to be created first
            """
            if self.app != "tables" or self.model != "row":
                raise RuntimeError("Error: This method is only available for recording objects.")
            # step1GetPresignedPostURL
            presigned = self.instance.api.get(self.instance.api.get_detail_url(self.app, self.model, sid=sid, query="presigned_post=true"))
            # step2UploadFileToS3
            if not "presigned_post" in presigned:
                raise RuntimeError("Error: File can not be uploaded! Make sure the sid is correct and you have permission to change it.")
            data = presigned["presigned_post"]["fields"]
            file_ob = open(filename, 'rb') # pylint: disable=consider-using-with
            #data["file"] = file_ob
            response = requests.post(url=presigned["presigned_post"]["url"], data=data, files={'file': file_ob})
            # step3UpdateVersionId
            if response.status_code != 204:
                raise RuntimeError(response.text)
            file_record = self.instance.api.get(self.instance.api.get_detail_url(self.app, self.model, sid=sid))
            # update file path
            cell_path_index = [index for index, cell in enumerate(file_record["column_set"]) if cell["column"]["widget"]["sid"] == "JMPS0a40x5eLQV16afk"][0]
            file_record["column_set"][cell_path_index] = self.instance.api.patch(
                self.instance.api.get_detail_url("tables", "cell", sid=file_record["column_set"][cell_path_index]["sid"]),
                {"data": f"{presigned['presigned_post']['fields']['key'].split('?')[0]}?versionId={response.headers['x-amz-version-id']}"}
            )
            # update file size
            cell_size_index = [index for index, cell in enumerate(file_record["column_set"]) if cell["column"]["widget"]["sid"] == "KNQT0a40x5fMRW27bgl"][0]
            file_record["column_set"][cell_size_index] = self.instance.api.patch(
                self.instance.api.get_detail_url("tables", "cell", sid=file_record["column_set"][cell_size_index]["sid"]),
                {"data": os.path.getsize(filename)}
            )
            return file_record

        def watch_folder(self, folder_path="", project__sids="", interval=5):
            """
                watch the a folder and upload files if found
                files will be uploaded to the files table
                after file is uploaded, it will be moved to a subfolder "uploaded"
                Args:
                    - folder_path, the folder or path to search
                    - project__sids, list of project sids
                    - interval, how often to check
            """
            if self.app != "tables" or self.model != "row":
                raise RuntimeError("Error: This method is only available for recording objects.")
            # default folder
            if folder_path == "":
                folder_path = "./"
            # add "/" to path
            if not folder_path.endswith("/"):
                folder_path = f"{folder_path}/"
            # create a uploaded folder
            if not os.path.isdir(f"{folder_path}/uploaded/"):
                os.system(f"mkdir -p {folder_path}/uploaded/")
            # get project
            if project__sids == "":
                project__sids = input("What project(s) do you wish to upload the file to? Provide the project SIDs separated by a ',': ")
            if project__sids.strip() == "":
                raise RuntimeError("Error: Invalid project SIDs!")
            # get files table
            print("Retrieving file table...")
            tables = self.instance.Table.list(serializer="name", query="name_system=file")
            if tables["count"] == 0:
                raise RuntimeError("Error: No file table is available. Create one and try again!")
            table_file = tables["results"][0]
            # get the columns
            columns = self.instance.Column.list(serializer="name", query=f"table__sid={table_file['sid']}")
            column_path = ""
            column_size = ""
            for column in columns["results"]:
                if column["name"] == "path":
                    column_path = column
                if column["name"] == "size":
                    column_size = column
            if column_path == "" or column_size == "":
                raise RuntimeError("Error: File path or size column is invalid!")
            # start watching
            print(f"Start watching folder ({folder_path})...")
            while True:
                time.sleep(interval)
                files = glob.glob(f'{folder_path}*')
                for file in files:
                    if not "/uploaded" in file:
                        # check token in case expired
                        self.instance.api.check_token()
                        # get file name
                        paths = file.split("/")
                        filename = paths[len(paths) - 1]
                        print(f"[{time.strftime('%Y-%m-%d %H:%M:%S')}] Uploading {filename}...", end="")
                        # create a file row
                        data_record = {
                            "name": filename,
                            "projects": [{"sid": sid.strip()} for sid in project__sids.split(",")],
                        }
                        data_record[column_path['sid']] = filename
                        data_record[column_size['sid']] = 0
                        record = self.instance.Record.create(data_record, query=f"table__sid={table_file['sid']}")
                        # upload the file
                        self.upload(record["sid"], file)
                        # move the file
                        os.system(f"mv {file} {folder_path}/uploaded/")
                        print("SUCCESS")



""" api related functions """
import json
import getpass
import requests

class APIObject:
    """ api object for api functions """
    base_url = "https://www.labii.dev"
    email = None
    password = None
    token = None
    organization__sid = None
    user = None

    def __init__(self, base_url="https://www.labii.dev", email=None, password=None, token=None, organization__sid=None, user=None):
        self.base_url = base_url
        self.email = email
        self.password = password
        self.token = token
        self.organization__sid = organization__sid
        self.user = user

    ######
    # url
    def get_list_url(self, app, model, level="organization", sid=None, serializer='list', query=""):
        """ return the list url """
        if sid is None:
            sid = self.organization__sid
        return f"/{app}/{model}/list/{level}/{sid}/{serializer}/{'' if query == '' else '?'}{query}"

    def get_detail_url(self, app, model, sid, query=""):
        """ return the detail url """
        return f"/{app}/{model}/detail/{sid}/{'' if query == '' else '?'}{query}"

    ######
    # auth
    def get_headers(self, is_authorized=True):
        """Return necessary headers for API functions

        Returns:
            dict: header objects
        """
        headers = {
            "X-Forwarded-For": "batch",
            "Content-Type": "application/json"
        }
        if is_authorized and self.token is not None:
            headers["Authorization"] = f"Token {self.token}"
        return headers

    def login(self, email="", password=""):
        """Get authentication token based on user input
        - self.login()
            - login with saved email/password
            - get email/password if not saved
        - self.login(email=xxx, password=xxx): login with new email/password

        Args:
            email (str): email to use to login
            password (str): password to use to login
        """
        # get email or password
        if email != "" and password != "":
            self.email = email
            self.password = password
        else:
            if self.email is None or self.password is None:
                self.email = input("Email: ")
                self.password = getpass.getpass('Password: ')
        # get token
        data = {
            "username": self.email,
            "password": self.password
        }
        response = self.post("/accounts/auth/", data, is_authorized=False)
        if "token" in response:
            self.token = response["token"]
            self.user = response
        elif "secret" in response:# two step authentication
            # use email to verify
            self.post("/accounts/mfa/code/", {"option": "email", "secret": response["secret"]}, is_authorized=False)
            code = input("MFA enabled for the account. Please check your email and provide authentication code: ")
            data = self.post("/accounts/mfa/verify/", {"code": code, "secret": response["secret"]}, is_authorized=False)
            if "token" in data:
                self.token = data["token"]
                self.user = data
            else:
                raise RuntimeError(data)
        else:
            raise RuntimeError(response)

    def check_token(self):
        """Check if the exist token is valid. If not valid, it will genereate a new token
            Labii token expires after 30 minutes of no activity. If your program take more than 30 minutes to run. Use this function to get a new token.
        """
        try:
            data = self.get("/accounts/checktoken/")
        except:
            #if data["detail"] != "Valid token.":
            self.login()

    ######
    # api post
    def post(self, url, data, is_authorized=True):
        """A function to do POST for labii api

        Args:
            url (str): API URL
            data (dict): data to POST, see API documentation (https://docs.labii.com/api/overview) for the format
            is_authorized (bool): should use the authrized token

        Returns:
            dict: created object
        """
        response = requests.post(url=f"{self.base_url}{url}", data=json.dumps(data, separators=(',',':')), headers=self.get_headers(is_authorized))
        try:
            data = json.loads(response.text)
            return data
        except:
            raise RuntimeError(response)

    def patch(self, url, data, is_authorized=True):
        """A function to do PATCH for labii api

        Args:
            url (str): API URL
            data (dict): data to update, see API documentation (https://docs.labii.com/api/overview) for the format
            is_authorized (bool): should use the authrized token

        Returns:
            dict: return of the api
        """
        response = requests.patch(url=f"{self.base_url}{url}", data=json.dumps(data, separators=(',',':')), headers=self.get_headers(is_authorized))
        try:
            data = json.loads(response.text)
            return data
        except:
            raise RuntimeError(response)

    def get(self, url, all_pages=False, is_authorized=True):
        """A function to do GET for labii api

        Args:
            url (str): API URL
            all (bool): should download all
            is_authorized (bool): should use the authrized token

        Returns:
            dict: result of the url
        """
        if all_pages is False:
            response = requests.get(url=f"{self.base_url}{url}", headers=self.get_headers(is_authorized))
            if response.status_code != 200:
                raise RuntimeError(response.text)
            try:
                data = json.loads(response.text)
                return data
            except:
                raise RuntimeError(response)
        else:
            page = 1
            page_size = 10
            # update url
            if "page" in url or "page_size" in url:
                query = url.split("?")[1].split("&")
                query_updated = []
                for item in query:
                    if not "page" in item and not "page_size" in item:
                        query_updated.append(item)
                if len(query_updated) > 0:
                    url = f"{url.split('?')[0]}?{'&'.join(query_updated)}"
                else:
                    url = url.split("?")[0]
            if "?" in url:
                url_with_page = f"{url}&page={page}&page_size={page_size}"
            else:
                url_with_page = f"{url}?page={page}&page_size={page_size}"
            response = requests.get(url=f"{self.base_url}{url_with_page}", headers=self.get_headers(is_authorized))
            if response.status_code != 200:
                raise RuntimeError(response.text)
            data = json.loads(response.text)
            count = data["count"]
            results = data["results"]
            print(f"Total records: {count}")
            print("Downloading page 1...")
            while len(results) < count:
                page += 1
                print(f"Downloading page {page}...")
                if "?" in url:
                    url_with_page = f"{url}&page={page}&page_size={page_size}"
                else:
                    url_with_page = f"{url}?page={page}&page_size={page_size}"
                response = requests.get(url=f"{self.base_url}{url_with_page}", headers=self.get_headers(is_authorized))
                results = results + json.loads(response.text)["results"]
            data["results"] = results
            return data

    def delete(self, url, is_authorized=True):
        """A function to do delete for labii api

        Args:
            url (str): API URL
            is_authorized (bool): should use the authrized token

        Returns:
            dict: result of the url
        """
        response = requests.delete(url=f"{self.base_url}{url}", headers=self.get_headers(is_authorized))
        if response.status_code != 200:
            raise RuntimeError(response.text)

#
# def labii_json_list_to_tsv(results):
#     """conver the list result in json format into tsv format
#
#     Args:
#         results (array): the results from response
#
#     Returns:
#         string: tsv data
#     """
#     if len(results) > 0:
#         should_collect_title = True
#         titles = [] # collect list of titles
#         data = [] # the data of all
#         for r in results:
#             d = [] # the data of one row
#             for field in r:
#                 if field == "column_set":
#                     for f in r["column_set"]:
#                         d.append(str(f["data"]).replace("\n","").replace("\r",""))
#                         if should_collect_title:
#                             titles.append(f["column"]["name"])
#                 elif field != "section_set":
#                     d.append(str(r[field]).replace("\n","").replace("\r",""))
#                     if should_collect_title:
#                         titles.append(field)
#             if should_collect_title:
#                 data.append("\t".join(titles))
#                 should_collect_title = False
#             data.append("\t".join(d))
#         return "\n".join(data)
#     else:
#         return "";
